from dataclasses import dataclass
from typing import List
from abc import ABC


@dataclass
class AbstractConfig(ABC):
    name: str
    reshaper: List[str]
    plotter: List[str]
    filename: str
    experiment_class = None  # type: ignore
    experiment_module: str



import json
import re
import sys
import argparse
import time
from typing import List
from copy import deepcopy
from naveen.data.pipe_cleaner import PipeCleaner
from naveen.data.scrapers.html_extractor import HTMLExtractor
from naveen.data.scrapers.html_tag_stripper import HTMLTagStripper
from bs4 import BeautifulSoup
from naveen.experiment.config.dynamic_config_maker import DynamicConfigMaker
from naveen.experiment.experiment_runner import ExperimentRunner  # type: ignore # noqa: E501

from naveen.utils.ngram_maker import NgramMaker


def pages() -> None:
    '''
    Input is assumed to be a url list

    An exclusion list of urls is required
    '''
    import argparse
    from naveen.data.page_maker import PageMaker

    parser = argparse.ArgumentParser()
    parser.add_argument('-b',
                        '--b',
                        '-base',
                        '--base',
                        '--baseurl',
                        type=str,
                        dest="baseurl",
                        required=True)
    parser.add_argument('-m',
                        '--m',
                        '-max',
                        '--max',
                        '--max_pages',
                        type=int,
                        dest="max_pages",
                        required=True)
    args = parser.parse_args()
    urls: List[str] = PageMaker.make_pages(args.baseurl, args.max_pages)
    for url in urls:
        print(url)


def filter() -> None:
    '''
    Input is assumed to be a url list

    An exclusion list of urls is required
    '''
    import argparse
    import sys
    from typing import Set

    if sys.stdin.isatty():  # no pipe
        print("- You must pipe input to filter")
        import os
        os._exit(0)

    parser = argparse.ArgumentParser()
    parser.add_argument('-e',
                        '--e',
                        '-exclude',
                        '--exclude',
                        type=str,
                        dest="excluded",
                        required=True)
    args = parser.parse_args()

    # assume args.excluded is a text file with urls
    with open(args.excluded, "r") as inf:
        excluded_urls: Set[str] = set(o.replace("\n", "") for o in inf)

    for line in sys.stdin:
        url: str = line  # type: ignore
        url = url.replace("\n", "")
        if url in excluded_urls:
            pass
        else:
            print(url)


def stream() -> None:
    from naveen.data.url_streamer import stream_url
    import sys

    if sys.stdin.isatty():  # no pipe
        print("- You must pipe input to stream")
        import os
        os._exit(0)

    for line in sys.stdin:  # for line in pipe
        line = line.replace('\n', '')
        stream_url(line)


def urls() -> None:
    import sys
    # https://www.geeksforgeeks.org/python-check-url-string/
    regex = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"  # noqa: E501

    for line in sys.stdin:
        urls = re.findall(regex, line)
        for url in urls:
            url = url[0].rstrip("\\").strip()
            url = url.replace('\n', '')
            if "," in url:
                url = url.split(",")[0]
            if '"' in url:
                url = url.split('"')[0]
            if url[-1] == "\\":
                url = url[0:-1]
            print(url)


def html2text() -> None:
    '''take raw html and turn it into a string'''

    for line_str in sys.stdin:
        line: dict = json.loads(line_str)
        soup = BeautifulSoup(line["html"], 'lxml')
        line["html"] = soup.text
        print(json.dumps(line))


def remove() -> None:
    import argparse

    import json
    msg = "Strip an html element from the dom"
    parser = argparse.ArgumentParser(msg)

    parser.add_argument("-e", "--e", '-element', "--element",
                        default="div",
                        dest="element", type=str,
                        required=True)

    args = parser.parse_args()

    for line_str in sys.stdin:
        line: dict = json.loads(line_str)
        html = line['html']  # type: ignore
        tag_stripper = HTMLTagStripper(html, args.element)
        line["html"] = tag_stripper.strip()
        print(json.dumps(line))


def pluck() -> None:
    import argparse
    import sys
    import json
    msg = "Pluck html elements with certain properties from a dom"
    parser = argparse.ArgumentParser(msg)
    parser.add_argument("-e", "--e", '-element', "--element",
                        default="div",
                        dest="element", type=str,
                        required=True)
    parser.add_argument("-a", "--a", '-attr', "--attribute",
                        dest="attribute", type=str,
                        default="class", required=True)
    parser.add_argument("-v", "--v", '-value', "--value",
                        dest="value", type=str,
                        default="class_name", required=True)
    args = parser.parse_args()

    for line_str in sys.stdin:
        line: dict = json.loads(line_str)
        assert "html" in line
        assert "url" in line
        extractor: HTMLExtractor = HTMLExtractor(line['html'])  # type: ignore
        messages = extractor.select_by_attr(attr=args.attribute,
                                            html_element=args.element,
                                            attr_value=args.value)
        for i, message in enumerate(messages):
            out: dict = deepcopy(line)
            out['html'] = str(message)
            print(json.dumps(out))


def clean() -> None:
    # 'Simple utility that returns a file name if the extension is added
    parser = argparse.ArgumentParser(description="skip steps in pipe")

    parser.add_argument('-e', '--e', '--extension',
                        dest="extension",
                        default=None,
                        required=True,
                        help="extension; include the dot")
    parser.add_argument('-o', '-output_dir',
                        '--output_dir',
                        dest="output_dir",
                        default=None,
                        required=True,
                        help="output directory")

    parser.add_argument('-v',
                        '--verbose',
                        dest="verbose",
                        default=False,
                        action=argparse.BooleanOptionalAction)  # type: ignore

    args = parser.parse_args()

    pipe_cleaner: PipeCleaner = PipeCleaner(
        args.extension,
        args.output_dir,
        args.verbose)

    '''
    # expected use
    # alias pipecleaner="python src/data/pipe_cleaner.py"
    # echo "test/fixtures/config/voss" | clean -e ".gz" =>
    # test/fixtures/config/voss.gz does not exist so print filename
    # echo "test/fixtures/config/voss" | clean -e ".json", prints "already exists"  # noqa: E501
    '''

    if sys.stdin.isatty():  # no pipe
        print("- You must pipe input to this program")
    else:
        for line in sys.stdin:  # for line in pipe
            line = line.replace('\n', '')
            already_exists: bool = pipe_cleaner.already_exists(line)
            if not already_exists:
                sys.stdout.write(line + "\n")
            else:
                # note writes to stderrr
                proposed = pipe_cleaner.get_proposed_filename(line)
                if args.verbose:
                    sys.stderr.write(
                        "[*] Already got {} => {}\n".format(line, proposed))


def ngrams() -> None:
    parser = argparse.ArgumentParser()
    help = "print all n-grams in a file to stdout; assume whitespace delimited"
    parser.add_argument("-n",
                        help=help,
                        type=int,
                        required=True)
    args = parser.parse_args()
    ngram_maker = NgramMaker(args.n)

    if sys.stdin.isatty():  # no pipe
        print("- You must pipe input to ngrams")
        import os
        os._exit(0)

    for line in sys.stdin:  # for line in pipe
        line = line.replace('\n', '')
        ngrams = ngram_maker.get_ngrams_from_whitespace_delimited_text(line)
        for n in ngrams:
            print(n)


def experiment() -> None:
    parser = argparse.ArgumentParser()
    help = "Run an experiment from a config file"
    parser.add_argument("-c", "--config",
                        help=help,
                        type=str,
                        dest="config",
                        required=True)
    parser.add_argument("-p", "--package_name",
                        help=help,
                        type=str,
                        dest="package_name",
                        required=True)
    parser.add_argument("-o", "--outputdir",
                        type=str,
                        dest="outputdir",
                        default="results")
    args = parser.parse_args()

    sys.path.append(args.package_name)

    maker = DynamicConfigMaker()
    msg = "Your config is {}. ".format(args.config)
    msg += "For now we assume config is json"
    assert args.config[-5:] == ".json", msg
    config = maker.from_json(args.config)
    name = args.outputdir + "/" + config.name + str(int(time.time()))
    runner = ExperimentRunner(config,
                              package_name=args.package_name,
                              output_directory=name)
    runner.run()


if __name__ == "__main__":
    a = 5
    print(a)



from abc import ABC
from abc import abstractmethod
from typing import List
from naveen.experiment.config.abstract_config import AbstractConfig  # type: ignore # noqa: E501


class Experiment(ABC):

    def __init__(self, config: AbstractConfig):
        self.config = config

    @abstractmethod
    def get_results(self) -> List[dict]:
        pass


class DemoExperiment(Experiment):
    # don't change this class of ir messes up tests
    def get_results(self) -> List[dict]:
        return [{"prod": "az", "hedonic": 8, "utilitatian": 9},
                {"prod": "aa", "hedonic": 9, "utilitatian": 9}]



import json


class ConfigValidator(object):

    def __init__(self) -> None:
        pass

    @staticmethod
    def is_valid_config_file(path_to_config: str) -> bool:
        with open(path_to_config, "r") as inf:
            jsonconfig: dict = json.load(inf)

            required_fields = ["name", "reshaper", "plotter",
                               "experiment_module", "experiment_class"]

            for field in required_fields:
                assert field in jsonconfig

        assert " " not in jsonconfig["name"], "Name cant have spaces"
        assert "/" not in jsonconfig["name"], "Name cant have slashes"

        return True



import importlib
import json
import unittest
import os
import glob

from dataclasses import fields
from naveen.data.scrapers.html_tag_stripper import HTMLTagStripper
from naveen.experiment.config.dynamic_config_maker import DynamicConfigMaker  # type: ignore # noqa: E501
from naveen.experiment.experiment_finisher import ExperimentFinisher
from naveen.experiment.experiment import DemoExperiment
from naveen.experiment.experiment_runner import ExperimentRunner
from naveen.experiment.experiment_writer import ExperimentWriter
from naveen.data.pipe_cleaner import PipeCleaner
from naveen.utils.ngram_maker import NgramMaker


def long_tests() -> None:
    os.system(
        "cat test/fixtures/url_list.txt | stream | wc -l > test/fixtures/output/stream2.txt")  # noqa: E501
    with open("test/fixtures/output/stream2.txt", "r") as inf:
        assert inf.read().strip() == "2"
    print("[*] Long tests passed")


class TestMethods(unittest.TestCase):

    def setUp(self) -> None:
        self.demo = 1
        maker = DynamicConfigMaker()
        self.config = maker.from_json("config/demo.json")
        for fn in glob.glob('test/fixtures/1659230924/*reshaped*'):
            os.remove(fn)
        for fn in glob.glob("test/fixtures/output/*"):
            os.remove(fn)

    def test_import_lib(self) -> None:
        module = importlib.import_module('src.experiment.experiment')
        my_class = getattr(module, 'DemoExperiment')
        my_instance = my_class(self.config)
        assert type(my_instance) == DemoExperiment

    def test_pipe_cleaner_true(self) -> None:
        cleaner: PipeCleaner = PipeCleaner(extension=".json", output_dir=".")
        exists = cleaner.already_exists("/tmp/a")
        self.assertFalse(exists)

    def test_pipe_cleaner_false(self) -> None:
        cleaner: PipeCleaner = PipeCleaner(
            extension=".json", output_dir="test/fixtures/config/")
        exists = cleaner.already_exists("test/fixtures/config/voss")
        self.assertTrue(exists)

    def test_demo(self) -> None:
        self.assertTrue(self.demo, 1)

    def test_demo_experiment_loads(self) -> None:
        demo_experiment = DemoExperiment(self.config)
        assert type(demo_experiment) == DemoExperiment

    def test_experiment_writer_loads(self) -> None:
        writer = ExperimentWriter(self.config)
        assert type(writer) == ExperimentWriter

    def test_experiment_runner(self) -> None:

        self.setUp()

        runner = ExperimentRunner(self.config,
                                  output_directory="test/fixtures/1659230924")

        runner.run()
        expected = "test/fixtures/1659230924/vossbooks.reshaped.csv"
        self.assertTrue(os.path.exists(expected))
        expected2 = "test/fixtures/1659230924/vossbooks.reshaped.pdf"
        self.assertTrue(os.path.exists(expected2))

    def test_exclusion_list(self) -> None:
        os.system("cat test/fixtures/url_list.txt | filter -e test/fixtures/exclusion_list.txt | wc -l > test/fixtures/output/exclusion1.txt")  # noqa: E501
        with open("test/fixtures/output/exclusion1.txt", "r") as inf:
            self.assertTrue(inf.read().strip(), "1")

    def test_dynamic_config_maker(self) -> None:
        maker = DynamicConfigMaker()
        config = maker.from_json("test/fixtures/config/voss.json")
        with open("test/fixtures/config/voss.json", "r") as inf:
            configdict = json.load(inf)
            names = [o.name for o in fields(config)]
            for k in configdict.keys():
                self.assertTrue(k in names)

    def test_finisher_reshaper(self) -> None:
        expected = "test/fixtures/1659230924"
        expected = expected + "/" + "vossbooks.reshaped.csv"
        if os.path.exists(expected):
            os.remove(expected)

        res = "test/fixtures/1659230924"
        finisher = ExperimentFinisher(self.config,
                                      results_directory=res)
        finisher.reshape()
        assert os.path.exists(expected)

    def test_finisher_plotter(self) -> None:
        expected = "test/fixtures/1659230924" + "/" + "vossbooks.reshaped.csv"
        if os.path.exists(expected):
            os.remove(expected)
        dir_ = "test/fixtures/1659230924"
        finisher = ExperimentFinisher(self.config, results_directory=dir_)
        finisher.reshape()
        finisher.plot()

    def test_tag_stripper(self) -> None:
        # https://allnurses.com/hedis-nursing-t709251/
        # should remove the block quotes
        with open("test/fixtures/blockquote.html", "r") as inf:
            html = inf.read()

        tag_stripper = HTMLTagStripper(html, "blockquote")

        quoted = "was thinking of taking the coarse you mentioned abov"
        self.assertTrue(quoted not in tag_stripper.strip())

    def test_ngrammer(self) -> None:

        line = "I am a fish"

        ngram_maker = NgramMaker(n=2)

        ngrams = ngram_maker.get_ngrams_from_whitespace_delimited_text(line)

        self.assertTrue("I am" in ngrams)
        self.assertTrue("am a" in ngrams)
        self.assertTrue("a fish" in ngrams)
        self.assertTrue(len(ngrams) == 3)

    def tearDown(self) -> None:
        os.system("rm -f Rplots.pdf")


if __name__ == '__main__':
    unittest.main()



'''
A utility for figuring out what can be skipped when piping lists
of file names. The assumption is that each step of the pipe adds
a new extension to the filename and may send the files to an
output directory
'''
import argparse
import sys
import os


class PipeCleaner(object):

    def __init__(self, extension: str,
                 output_dir: str,
                 verbose: bool = False) -> None:
        assert extension != "", (
            "You can't have an empty extension")
        assert output_dir is not None, (
            "You must specify and output dir")
        assert extension[0] == ".", (
            "Extension must start with a dot")
        assert extension != ".", (
            "You must have at least one character in extension")
        assert os.path.isdir(
            output_dir), "Must specify a valid output directory"
        self.extension = extension
        self.output_dir = output_dir
        self.verbose = verbose

    def get_proposed_filename(self, path_to: str) -> str:
        basename = os.path.basename(path_to)
        proposed_name = os.path.join(
            self.output_dir, basename + self.extension)
        if self.verbose:
            print("[*] proposed name {}".format(proposed_name))
        return proposed_name

    def already_exists(self, path_to: str) -> bool:
        proposed_name: str = self.get_proposed_filename(path_to)
        out: bool = os.path.isfile(proposed_name)
        if self.verbose:
            print("[*] proposed name {} exists? {}".format(proposed_name, out))
        return out


if __name__ == "__main__":
    # 'Simple utility that returns a file name if the extension is added
    parser = argparse.ArgumentParser(description="skip steps in pipe")

    parser.add_argument('-e', '--e', '--extension',
                        dest="extension",
                        default=None,
                        help="extension that piped process will add")
    parser.add_argument('-o', '-output_dir',
                        '--output_dir',
                        dest="output_dir",
                        default=None,
                        help="output directory")
    args = parser.parse_args()

    pipe_cleaner: PipeCleaner = PipeCleaner(args.extension, args.output_dir)

    '''
    # expected use
    # alias pipecleaner="python src/data/pipe_cleaner.py"
    # echo "test/fixtures/config/voss" | pipecleaner -e ".gz" =>
    # test/fixtures/config/voss.gz does not exist so print filename
    # echo "test/fixtures/config/voss" | pipecleaner -e ".json", prints "already exists"  # noqa: E501
    '''

    if sys.stdin.isatty():  # no pipe
        print("- You must pipe input to this program")
    else:
        for line in sys.stdin:  # for line in pipe
            line = line.replace('\n', '')
            already_exists: bool = pipe_cleaner.already_exists(line)
            if not already_exists:
                sys.stdout.write(line + "\n")
            else:
                # note writes to stderrr
                proposed = pipe_cleaner.get_proposed_filename(line)
                sys.stderr.write(
                    "[*] Already got {} => {}\n".format(line, proposed))



import subprocess
from naveen.experiment.config.abstract_config import AbstractConfig  # type: ignore # noqa: E501


class ExperimentFinisher(object):

    def __init__(self, config: AbstractConfig, results_directory: str):
        '''results directory is usually a unit timestamp'''
        self.config = config
        self.results_directory = results_directory

    def reshape(self) -> None:
        scriptargs = self.config.reshaper
        if self.config.reshaper == []:
            return None
        # e.g. ["python", "scripts/reshapers/demo.py"]
        command, script = scriptargs[0:2]
        assert script.startswith("scripts/reshaper/")
        subprocess.run(scriptargs + ["--results_directory", self.results_directory] + [  # noqa: E501
            "--results_file", self.config.name + ".csv"])

    def plot(self) -> None:

        scriptargs = self.config.plotter
        if self.config.plotter == []:
            return None
        # e.g. ["R", "scripts/plotters/demo.R"]
        command, script = scriptargs[0:2]
        assert script.startswith("scripts/plotter/")
        # for now assume that the plotter takes one arg,
        # which is self.config["name"] + ".reshaped.csv"
        subprocess.run(
            scriptargs + [self.results_directory + "/" + self.config.name + ".reshaped.csv"])  # noqa: E501

    def finish_results(self) -> None:
        self.reshape()
        self.plot()



import argparse
import sys
from typing import List


class NgramMaker(object):

    def __init__(self, n: int = 5) -> None:
        self.n = n

    def get_ngrams_from_whitespace_delimited_text(self,
                                                  text: str) -> List[str]:
        words = text.split()
        out: List[str] = []
        for gm in zip(*[words[i:] for i in range(self.n)]):
            gram = " ".join(gm)
            out.append(gram)
        return out


if __name__ == "__main`__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-n", help="how many grams",
                        type=int,
                        default=7)
    args = parser.parse_args()
    ngram_maker = NgramMaker(args.n)

    if sys.stdin.isatty():  # no pipe
        print("- You must pipe input to ngrams")
        import os
        os._exit(0)

    for line in sys.stdin:  # for line in pipe
        line = line.replace('\n', '')
        ngrams = ngram_maker.get_ngrams_from_whitespace_delimited_text(line)
        print(ngrams)
        for n in ngrams:
            print(n)



from dataclasses import make_dataclass
from typing import List
import json
from typing import Tuple
from typing import Any

from naveen.experiment.config.config_validator import ConfigValidator


class DynamicConfigMaker(object):

    # figure out type anno here is too hard
    def from_json(self, config_file: str):  # type: ignore

        validator: ConfigValidator = ConfigValidator()
        assert validator.is_valid_config_file(config_file)

        with open(config_file, "r") as inf:
            jsonconfig = json.load(inf)
            jsonconfig['filename'] = config_file

        # https://stackoverflow.com/questions/52534427/dynamically-add-fields-to-dataclass-objects
        fields: List[Tuple[str, Any]] = []
        fields.append(("name", str))
        fields.append(("reshaper", List[str]))
        fields.append(("plotter", List[str]))
        for k, v in jsonconfig.items():
            if k not in ["name", "reshaper", "plotter"]:
                fields.append((k, type(v)))

        X = make_dataclass('DynamicConfig', fields)

        # https://www.reddit.com/r/learnpython/comments/9h74no/convert_dict_to_dataclass/
        return X(**jsonconfig)



'''
Stream a url piped in
'''
import sys
import json
import subprocess
from naveen.logger import get_logger


def stream_url(url: str, timeout: int = 3) -> None:
    '''time out unit is seconds, not ms!'''

    out = subprocess.run(["wget", url, "-q",
                          "--timeout={}".format(str(timeout)),
                          "-O", "-"],
                         capture_output=True)
    # returns std out as a unicode string, "return code is wget return code"
    # https://www.gnu.org/software/wget/manual/html_node/Exit-Status.html
    # exit code 0 means no problem
    logger = get_logger()
    try:
        print(json.dumps({"url": url,
                          "returncode": out.returncode,
                          "html": out.stdout.decode("utf-8")}))
    except UnicodeDecodeError:
        logger.error(
            "[*] Error, could not decode {} and so skipping it".format(url))


if __name__ == "__main__":

    url = sys.argv[1]
    url = url.replace('\n', '')
    stream_url(url)
